/*
 * hp.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"
#include "hptoken.h"
#include "hptsb.h"
#include "decl.h"

uint8_t	line[32768];

/*
 * HP-2000/ACCESS Directory Entry.
 *	Word	Bytes	Name	Usage
 *	0	0-1	USERID	User ID
 *	1	2-3	FILNM1	Program or File Name bytes 1-2.
 *	2	3-4	FILNM2	Program or File Name bytes 3-4.
 *	3	5-6	FILNM3	Program or File Name bytes 5-6.
 *				Names are left-justified, space (0x20)
 *				
 */

void
eval_2000access_equipment_table (uint8_t *buf, uint32_t rlen)
{
	int	i;

	eqAC.NIDT		= WTOI (buf);
	eqAC.NDIRT		= WTOI (buf + 2);
	for (i = 0; i < 8; i++) {
		eqAC.ADTAT[i * 3 + 0] = WTOI (buf + 002 * 2 + (i * 3) * 2);
		eqAC.ADTAT[i * 3 + 1] = WTOI (buf + 002 * 2 + (i * 3) * 2 + 2);
		eqAC.ADTAT[i * 3 + 2] = WTOI (buf + 002 * 2 + (i * 3) * 2 + 4);
	}
	for (i = 0; i < 8; i++)
		eqAC.DKTBL[i] = WTOI (buf + 032 * 2 + i * 2);
	eqAC.DADSL[0]		= WTOI (buf + 042 * 2);
	eqAC.DADSL[1]		= WTOI (buf + 043 * 2);
	eqAC.DLNSL		= WTOI (buf + 044 * 2);
	memcpy (eqAC.SYSID, buf + 045 * 2, 10);
	eqAC.MAGSC		= WTOI (buf + 052 * 2);
	eqAC.NPORT		= 65536 - WTOI (buf + 053 * 2);
	eqAC.YEAR		= WTOI (buf + 054 * 2);
	eqAC.HOUR		= WTOI (buf + 055 * 2);
	eqAC.TENTHS		= WTOI (buf + 056 * 2);
	eqAC.HDATE		= WTOI (buf + 057 * 2);
	eqAC.SLEPT		= WTOI (buf + 060 * 2);

	dprintf (DEBUG_HEADERS, "2000/ACCESS Equipment table:\n");
	dprintf (DEBUG_HEADERS, "    Number of ID tracks:                 %d\n", eqAC.NIDT);
	dprintf (DEBUG_HEADERS, "    Directory tracks per disc:           %d\n", eqAC.NDIRT);
	dprintf (DEBUG_HEADERS, "    ADT Disc Address Table:\n");
	for (i = 0; i < 8; i++)
		if (eqAC.ADTAT[i * 3 + 2])
			dprintf (DEBUG_HEADERS, "        %d:  Cylinder: %d  Head: %d  Sector: %d\n",
					i,
					eqAC.ADTAT[i * 3 + 0] & 0xFF,
					(eqAC.ADTAT[i * 3 + 1] >> 8) & 0x1F,
					eqAC.ADTAT[i * 3 + 2] & 0xFF);
					
	dprintf (DEBUG_HEADERS, "    Disc select code/unit numbers:\n");
	for (i = 0; i < 8; i++)
		if (eqAC.DKTBL[i] != 0)
			dprintf (DEBUG_HEADERS, "        %d:  Select code %d, Unit Number %d\n",
					i, eqAC.DKTBL[i] >> 8, eqAC.DKTBL[i] & 0377);
	dprintf (DEBUG_HEADERS, "    Disc Address of System Library:       Select Code %d, Unit Number %d, Cylinder %d, Head %d, Sector %d\n",
			eqAC.DKTBL[eqAC.DADSL[0] & 0xFF] >> 8,
			eqAC.DKTBL[eqAC.DADSL[0] & 0xFF] & 0xFF,
			eqAC.DADSL[0] & 0xFF,
			eqAC.DADSL[1] & 0xFF,
			(eqAC.DADSL[1] >> 8) & 0x1F,
			eqAC.DADSL[1] & 0xFF);
	dprintf (DEBUG_HEADERS, "    Length of System library (blocks):    %d\n", eqAC.DLNSL);
	dprintf (DEBUG_HEADERS, "    System ID:                            %10.10s\n", eqAC.SYSID);
	dprintf (DEBUG_HEADERS, "    Mag Tape Select Code/Terminet:        %o, Terminet=%s\n", eqAC.MAGSC & 0x7fffffff,
			eqAC.MAGSC & 0x80000000 ? "Yes" : "No");
	dprintf (DEBUG_HEADERS, "    Number of ports:                      %u\n", eqAC.NPORT);
	dprintf (DEBUG_HEADERS, "    Tape Generation Date:                 ");
	if (verbose > 2)
		day2date (eqAC.YEAR, eqAC.HOUR, eqAC.TENTHS);
	dprintf (DEBUG_HEADERS, "  [%d/%d/%d]\n", eqAC.YEAR, eqAC.YEAR, eqAC.HOUR, eqAC.TENTHS);
	dprintf (DEBUG_HEADERS, "    Tape Generation HIB Date:             %d\n", eqAC.HDATE);
	dprintf (DEBUG_HEADERS, "    Tape Slept?                           %s\n", eqAC.SLEPT ? "No" : "Yes");
	return;
}

void
eval_2000C_equipment_table (uint8_t *buf, uint32_t rlen)
{
	uint8_t	*	pbuf = buf + 2;
	int		i;

	for (i = 0; i < 3; i++) {
		eqC.IDEC[i]= WTOI (pbuf + 2 + i * 2);
		eqC.IDEC[0]		= WTOI (pbuf + (i * 3) + 0);	/* First ID of track */
		eqC.IDEC[1]		= WTOI (pbuf + (i * 3) + 2);	/* Drum address of track */
		eqC.IDEC[2]		= WTOI (pbuf + (i * 3) + 4);	/* Length in -words of track */
	}
	eqC.NIDT		= WTOI (pbuf + 011 * 2);
	eqC.ADLOC		= WTOI (pbuf + 012 * 2);
	eqC.ADLEN		= WTOI (pbuf + 013 * 2);
	eqC.NDIRT		= WTOI (pbuf + 014 * 2);
	for (i = 0; i < 8; i++)
		eqC.DADLC[i] = WTOI (pbuf + (015 + i) * 2);
	for (i = 0; i < 8; i++)
		eqC.DADLN[i] = WTOI (pbuf + (025 + i) * 2);
	for (i = 0; i < 4; i++)
		eqC.DTBL[i] = WTOI (pbuf + (035 + i) * 2);
	for (i = 0; i < 8; i++)
		eqC.DKTBL[i] = WTOI (pbuf + (041 + i) * 2);
	for (i = 0; i < 16; i++)
		eqC.TRAX[i] = WTOI (pbuf + (051 + i) * 2);
	memcpy (eqC.SYSID, pbuf + 071 * 2, 10); /* Was 071 */
	eqC.MAGSC		= WTOI (pbuf + 076 * 2);
	eqC.NPORT		= 65536 - WTOI (pbuf + 077 * 2);
	eqC.YEAR		= WTOI (pbuf + 0100 * 2);
	eqC.HOUR		= WTOI (pbuf + 0101 * 2);
	eqC.TENTHS		= WTOI (pbuf + 0102 * 2);
	eqC.HDATE		= WTOI (pbuf + 0103 * 2);
	eqC.SLEPT		= WTOI (pbuf + 0104 * 2);

	dprintf (DEBUG_HEADERS, "2000C Equipment table:\n");
	dprintf (DEBUG_HEADERS, "    ID Table Headers:\n");
	for (i = 0; i < 3; i++) {
		if (eqC.IDEC[i * 4 + 3] != 0)
			dprintf (DEBUG_HEADERS, "        %d: ID: %c%03d  Disc Addr: %06d/%06d  Len: %d\n",
					i,
					eqC.IDEC[i * 4] / 1024 + '@', eqC.IDEC[i * 4] % 1024,
					eqC.IDEC[i * 4 + 1], eqC.IDEC[i * 4 + 2],
					eqC.IDEC[i * 4 + 3]);
	}
	dprintf (DEBUG_HEADERS, "    Number of ID tracks:                 %d\n", eqC.NIDT);
	dprintf (DEBUG_HEADERS, "    Drum address of ADT:                 %o\n", eqC.ADLOC);
	dprintf (DEBUG_HEADERS, "    Length of ADT in blocks:             %d\n", 65536 - eqC.ADLEN);
	dprintf (DEBUG_HEADERS, "    Directory tracks per disc:           %d\n", eqC.NDIRT);
	dprintf (DEBUG_HEADERS, "    Drum addresses of Disc ADT tracks:\n");
	for (i = 0; i < 8; i++)
		dprintf (DEBUG_HEADERS, "        %d:  %06o   Len: %dn",
				i, eqC.DADLC[i], 65536 - eqC.DADLN[i]);
	dprintf (DEBUG_HEADERS, "    Drum Existence Table:\n");
	for (i = 0; i < 4; i++)
		dprintf (DEBUG_HEADERS, "        %d:  Prefix: %o  Select Code: %o\n",
				i, (eqC.DTBL[i] >> 6) & 03, eqC.DTBL[i] & 077);
	dprintf (DEBUG_HEADERS, "    Disc Existence Table:\n");
	for (i = 0; i < 8; i++)
		dprintf (DEBUG_HEADERS, "        %d:  Select Code: %o  Unit Number: %o\n",
				i, eqC.DKTBL[i] >> 8, eqC.DKTBL[i] & 0377);
	dprintf (DEBUG_HEADERS, "    Drum Track Availability (TRAX) Table:\n");
	for (i = 0; i < 4; i++)
		dprintf (DEBUG_HEADERS, "        Drum %d: %06o %06o %06o %06o\n",
				i, eqC.TRAX[i * 4], eqC.TRAX[i * 4 + 1],
				eqC.TRAX[i * 4 + 2], eqC.TRAX[i * 4 + 3]);
	dprintf (DEBUG_HEADERS, "    System ID:                           %.10s\n", eqC.SYSID);
	dprintf (DEBUG_HEADERS, "    Mag Tape Select Code:                %o\n", eqC.MAGSC & 0x7fffffff);
	dprintf (DEBUG_HEADERS, "    Terminet?                            %s\n", eqC.MAGSC & 0x80000000 ? "Yes" : "No");
	dprintf (DEBUG_HEADERS, "    Number of terminal ports:            %u\n", eqC.NPORT);
	dprintf (DEBUG_HEADERS, "    Tape Generation Date:                ");
	if (verbose > 2)
		day2date (eqC.YEAR, eqC.HOUR, eqC.TENTHS);
	dprintf (DEBUG_HEADERS, "  [%d/%d/%d]\n", eqC.YEAR, eqC.HOUR, eqC.TENTHS);
	dprintf (DEBUG_HEADERS, "    Tape Generation HIB Date:            %d\n", eqC.HDATE);
	dprintf (DEBUG_HEADERS, "    Tape Slept?:                         %s\n", eqC.SLEPT ? "No" : "Yes");
	return;
}

void
eval_2000F_equipment_table (uint8_t *buf, uint32_t rlen)
{
	uint8_t	*	pbuf = buf + 2;	/* Offset from the record length */
	int		i;

	for (i = 0; i < 3; i++) {
		eqF.IDEC[i * 4]		= WTOI (pbuf + (i * 4) * 2);
		eqF.IDEC[i * 4 + 1]	= WTOI (pbuf + (i * 4) * 2 + 2);
		eqF.IDEC[i * 4 + 2]	= WTOI (pbuf + (i * 4) * 2 + 4);
		eqF.IDEC[i * 4 + 3]	= WTOI (pbuf + (i * 4) * 2 + 6);
	}
	eqF.NIDT		= WTOI (pbuf + 014 * 2);
	eqF.NDIRT		= WTOI (pbuf + 015 * 2);
	for (i = 0; i < 8; i++) {
		eqF.ADTAT[i * 3 + 0] = WTOI (pbuf + 016 * 2 + (i * 3) * 2);
		eqF.ADTAT[i * 3 + 1] = WTOI (pbuf + 016 * 2 + (i * 3) * 2 + 2);
		eqF.ADTAT[i * 3 + 2] = WTOI (pbuf + 016 * 2 + (i * 3) * 2 + 4);
	}
	eqF.ADTLN		= WTOI (pbuf + 044 * 2);
	for (i = 0; i < 8; i++)
		eqF.DKTBL[i] = WTOI (pbuf + 046 * 2 + i * 2);
	eqF.DADSL[0]		= WTOI (pbuf + 056 * 2);
	eqF.DADSL[1]		= WTOI (pbuf + 057 * 2);
	eqF.DLNSL		= WTOI (pbuf + 060 * 2);
	memcpy (eqF.SYSID, pbuf + 061 * 2, 10);
	eqF.MAGSC		= WTOI (pbuf + 066 * 2);
	eqF.NPORT		= 65536 - WTOI (pbuf + 067 * 2);
	eqF.YEAR		= WTOI (pbuf + 070 * 2);
	eqF.HOUR		= WTOI (pbuf + 071 * 2);
	eqF.TENTHS		= WTOI (pbuf + 072 * 2);
	eqF.HDATE		= WTOI (pbuf + 073 * 2);
	eqF.SLEPT		= WTOI (pbuf + 074 * 2);
	dprintf (DEBUG_HEADERS, "2000F Equipment table:\n");
	dprintf (DEBUG_HEADERS, "    ID Table Headers:\n");
	for (i = 0; i < 3; i++) {
		if (eqF.IDEC[i * 4 + 3] != 0)
			dprintf (DEBUG_HEADERS, "        %d: ID: %c%03d  Disc Addr: %06d/%06d  Len: %d (%o)\n",
					i,
					eqF.IDEC[i * 4] / 1024 + '@', eqF.IDEC[i * 4] % 1024,
					eqF.IDEC[i * 4 + 1], eqF.IDEC[i * 4 + 2],
					65536 - eqF.IDEC[i * 4 + 3], eqF.IDEC[i * 4 + 3]);
	}
	dprintf (DEBUG_HEADERS, "    Number of ID tracks:                 %d\n", eqF.NIDT);
	dprintf (DEBUG_HEADERS, "    Directory tracks per disc:           %d\n", eqF.NDIRT);
	dprintf (DEBUG_HEADERS, "    ADT Disc Address Table:\n");
	for (i = 0; i < 8; i++)
		if (eqF.ADTAT[i * 3 + 2])
			dprintf (DEBUG_HEADERS, "        %d: Disc addr: %06d/%06d  Len: %d (%o)\n",
					i,
					eqF.ADTAT[i * 3 + 0], eqF.ADTAT[i * 3 + 1], 
					65536 - eqF.ADTAT[i * 3 + 2], eqF.ADTAT[i * 3 + 2]);
	dprintf (DEBUG_HEADERS, "    Disc select code/unit numbers:\n");
	for (i = 0; i < 8; i++)
		if (eqF.DKTBL[i] != 0)
			dprintf (DEBUG_HEADERS, "        %d: Select Code: %d   Unit Number: %d  (%06o)\n",
					i, eqF.DKTBL[i] >> 8, eqF.DKTBL[i] & 0377, eqF.DKTBL[i]);
	dprintf (DEBUG_HEADERS, "    System ID:                           %.10s\n", eqF.SYSID);
	dprintf (DEBUG_HEADERS, "    Mag Tape Select Code:                %o\n", eqF.MAGSC);
	dprintf (DEBUG_HEADERS, "    Number of terminal ports:            %u\n", eqF.NPORT);
	dprintf (DEBUG_HEADERS, "    Tape Generation Date:                ");
	if (verbose > 2)
		day2date (eqF.YEAR, eqF.HOUR, eqF.TENTHS);
	dprintf (DEBUG_HEADERS, "  [%d/%d/%d]\n", eqF.YEAR, eqF.HOUR, eqF.TENTHS);
	dprintf (DEBUG_HEADERS, "    Tape generation HIB Date:            %d\n", eqF.HDATE);
	dprintf (DEBUG_HEADERS, "    Tape slept?:                         %s\n", eqF.SLEPT ? "No" : "Yes");
	return;
}

void
eval_2000E_equipment_table (uint8_t *buf, uint32_t rlen)
{
	uint8_t	*	pbuf = buf + 2;
	int		i;

	/*
	 * 000 - 067 = DIREC	DIREC Table
	 *			  0	-length in words of first directory track of Disc 0
	 *			 1-4	First 4 words of first directory track of Disc 0
	 *			  5	unused
	 *			  6	Disc address of first directory track of Disc 0
	 *			 7-13	Same as 0-6, but for 2nd directory track of Disc 0
	 *			14-20		1st Directory, Disc 1
	 *			21-27		2nd Directory, Disc 1
	 *			28-34		1st Directory, Disc 2
	 *			35-41		2nd Directory, Disc 2
	 *			42-48		1st Directory, Disc 3
	 *			49-55		2nd Directory, Disc 3
	 *    070    = IDLEN	Length of ID table
	 *    071    = IDTTA	Disc Pointer to ID Track Addresses
	 * 072 - 075 = IDTAT	Disc Addresses of ID Tracks (4th track comes first in table)
	 *    076    = IDTRL	Disc Pointer to ID Track Lengths
	 * 077 - 102 = IDTLT	ID Track Length Table (4th track comes first in table)
	 *    103    = ADLOC	System Disc address of ADT
	 *    104    = ADLEN	-number of words in ADT
	 *    105    = SDT	Sectors per Disc Track
	 *    106    = NPORT	-number of swap tracks reserved.  2000E is always 16 port.
	 *    107    = PHR	Number of seconds for user to log on
	 *    108    = SLEPT	One if slept, zero if not.
	 */

	for (i = 0; i < 4; i++) {
		/*eqE.DIREC[i].len = WTOI (pbuf);*/
		eqE.DIREC[i].id = WTOI (pbuf + (i * 4 + 000) * 2);
		memcpy(eqE.DIREC[i].filnam, pbuf + (i * 4 + 001) * 2, 6);
		eqE.DIREC[i].address = WTOI (pbuf + (i * 4 + 004) * 2);
	}
	eqE.IDLEN = WTOI (pbuf + 070 * 2);
	eqE.IDTTA = WTOI (pbuf + 071 * 2);
	for (i = 0; i < 8; i++)
		eqE.IDTAT[i] = WTOI (pbuf + (072 + i) * 2);
	eqE.IDTRL = WTOI (pbuf + 076 * 2);
	for (i = 0; i < 8; i++)
		eqE.IDTLT[i] = WTOI (pbuf + (077 + i) * 2);
	eqE.ADLOC = WTOI (pbuf + 0103 * 2);
	eqE.ADLEN = WTOI (pbuf + 0104 * 2);
	eqE.SDT = WTOI (pbuf + 0105 * 2);
	eqE.NPORT = WTOI (pbuf + 0106 * 2);
	eqE.PHR = WTOI (pbuf + 0107 * 2);
	eqE.SLEPT = WTOI (pbuf + 0110 * 2);

	dprintf (DEBUG_HEADERS, "2000E Equipment table:\n");
	dprintf (DEBUG_HEADERS, "    DIREC Table:\n");
	for (i = 0; i < 4; i++) {
		/*if (eqC.IDEC[i * 4 + 3] != 0)*/
			dprintf (DEBUG_HEADERS, "        %d: ID: %c%03d  Filnam: %6.6s  %s Disc %d, Track %d, Sector %d\n",
					i,
					eqE.DIREC[i * 4].id / 1024 + '@', eqE.DIREC[i * 4].id % 1024,
					eqE.DIREC[i * 4].filnam,
					((eqE.DIREC[i * 4].address >> 6) & 0x01) ? "Removable" : "Fixed",
					eqE.DIREC[i * 4].address >> 15,
					((eqE.DIREC[i * 4].address) >> 7) & 0xff,
					eqE.DIREC[i * 4].address & 0x3f);
	}
	dprintf (DEBUG_HEADERS, "    ID Table Length:                     %d\n", 65536 - eqE.IDLEN);
	dprintf (DEBUG_HEADERS, "    Disc address of ID Track Table:      %o\n", eqE.IDTTA);
	dprintf (DEBUG_HEADERS, "    Disc Addresses of ID Tracks:\n");
	for (i = 0; i < 8; i++)
		dprintf (DEBUG_HEADERS, "        %d:  %06o\n", i, eqE.IDTAT[i]);
	dprintf (DEBUG_HEADERS, "    Disc addr. of ID Track Lengh Table:  %o\n", eqE.IDTRL);
	dprintf (DEBUG_HEADERS, "    Disc Addresses of ID Track Lengths:\n");
	for (i = 0; i < 4; i++)
		dprintf (DEBUG_HEADERS, "        %d:  %06o\n", i, eqE.IDTLT);
	dprintf (DEBUG_HEADERS, "    System Disc Address of ADT:          %o\n", eqE.ADLOC);
	dprintf (DEBUG_HEADERS, "    ADT length:                          %o\n", 65536 - eqE.ADLEN);
	dprintf (DEBUG_HEADERS, "    SDT - Sectors per Disc Track:        %o\n", eqE.SDT);
	dprintf (DEBUG_HEADERS, "    Number of Ports:                     %o\n", 65536 - eqE.NPORT);
	dprintf (DEBUG_HEADERS, "    Reserved Swap Tracks:                %o\n", 65536 - eqE.PHR);
	dprintf (DEBUG_HEADERS, "    Tape slept?:                         %s\n", eqE.SLEPT ? "No" : "Yes");
}

void
eval_ACCESS_DIREC_table (uint8_t *buf, uint32_t rlen)
{
	uint32_t	i;
	uint8_t		*j;

	dprintf (DEBUG_TABLES, "Processing 2000/ACCESS DIREC Table.\n");

	
	for (i = 0; i < 80; i++) {
		j = buf + (80 * i) + (7 * 2);
		Dt2000access.track[i].len = WTOI (j);
		Dt2000access.track[i].first_directory_userid = WTOI (j + 2);
		memcpy(Dt2000access.track[i].first_directory_name, j + 4, 6);
		Dt2000access.track[i].disc_address_1 = WTOI (j + 10);
		Dt2000access.track[i].disc_address_2 = WTOI (j + 12);
		dprintf (DEBUG_TABLES, "  Entry %2d: len=%6d\tuserid=%c%03d\tADDR1=%6d\tADDR2=%6d\n",
				i,
				65536 - Dt2000access.track[i].len, 
				(Dt2000access.track[i].first_directory_userid / 1024) + '@',
				Dt2000access.track[i].first_directory_userid & 0x3ff,
				Dt2000access.track[i].disc_address_1,
				Dt2000access.track[i].disc_address_2);
	}
	
}

void
eval_2000C_DIREC_table (uint8_t *buf, uint32_t rlen)
{
	uint32_t	i;
	uint8_t		*j;

	dprintf (DEBUG_TABLES, "Processing 2000C/2000F DIREC Table.\n");

	
	for (i = 0; i < 80; i++) {
		j = buf + i * 7 * 2 + 2;
		Dt2000access.track[i].len = WTOI (j);
		Dt2000access.track[i].first_directory_userid = WTOI (j + 1 * 2);
		memcpy(Dt2000access.track[i].first_directory_name, j + 2 * 2, 6);
		Dt2000access.track[i].disc_address_1 = WTOI (j + 10 * 2);
		Dt2000access.track[i].disc_address_2 = WTOI (j + 12 * 2);
		if (Dt2000access.track[i].len != 0) {
			dprintf (DEBUG_TABLES, "  Entry %2d: len=%6d\tuserid=%c%03d\tADDR=%06o\tADDR2=%06o\n",
				i,
				65536 - Dt2000access.track[i].len, 
				(Dt2000access.track[i].first_directory_userid / 1024) + '@',
				Dt2000access.track[i].first_directory_userid & 0x3ff,
				Dt2000access.track[i].disc_address_1,
				Dt2000access.track[i].disc_address_2);
		}
	}
	
}

void
eval_2000access_NONSHAREDEVICE_table (uint8_t *buf, uint32_t rlen)
{
	uint32_t	i;

		dprintf (DEBUG_TABLES, "Processing 2000/ACCESS NON_SHAREABLE DEVICE Table.\n");

	NSDEV2000access.ndevices = WTOI (buf);

	for (i = 0; i < 64; i++) {
		NSDEV2000access.entry[i].data1 = WTOI (buf + 1 + (i * 5));
		NSDEV2000access.entry[i].data2 = WTOI (buf + 1 + (i * 5) + 1);
		NSDEV2000access.entry[i].data3 = WTOI (buf + 1 + (i * 5) + 2);
		NSDEV2000access.entry[i].data4 = WTOI (buf + 1 + (i * 5) + 3);
		NSDEV2000access.entry[i].data5 = WTOI (buf + 1 + (i * 5) + 4);
	}
}

void
display_2000access_NONSHAREDEVICE_table (void)
{
	uint32_t	i,j;
#ifdef NOTYET
	uint8_t		assignment[5];
#endif

	dprintf (DEBUG_TABLES, "  NON-SHAREABLE DEVICE Table:\n");
	dprintf (DEBUG_TABLES, "\n  DEVICE    SELECT    MAXIMUM\n");
	dprintf (DEBUG_TABLES, "DESIGNATOR   CODE   RECORD SIZE  ASSIGNMENT    USER\n");
#ifdef NOTYET
	for (i = 0; i < (65536 - NSDEV2000access.ndevices); i++) {
		switch (NSDEV2000access.entry[i].data4) {
			case 1:		strcpy (assignment, "RJE");
					break;
			case 0:		strcpy (assignment, "ALL");
					break;
			case -1:	strcpy (assignment, "NONE");
					break;
			default:	sprintf (assignment, "%c%d", 
						NSDEV2000access.entry[i].data4 / 1024 + "@",
						NSDEV2000access.entry[i].data4 % 1024);
		}
		dprintf (DEBUG_TABLES, "  %c%c%d         %02o         %4d        %s       %d\n",
				((NSDEV2000access.entry[i].data1 >> 10) & 0x1f) + 0101,
				((NSDEV2000access.entry[i].data1 >> 5) & 0x1f) + 0101,
				(NSDEV2000access.entry[i].data1 & 0x1f),
				NSDEV2000access.entry[i].data2 & 0xff,
				NSDEV2000access.entry[i].data3,
				assignment,
				NSDEV2000access.entry[i].data5);
	}
#endif
	for (i = 0; i < 64; i++) {
	    for (j = 0; j < 5; j++) {
		dprintf (DEBUG_TABLES, "%d %c%c%d\n", i * 5 + j,
			((NSDEV2000access.entry[i * 5 + j].data1 >> 10) & 0x1f) + 0101,
			((NSDEV2000access.entry[i * 5 + j].data1 >> 5) & 0x1f) + 0101,
			(NSDEV2000access.entry[i * 5 + j].data1 & 0x1f));
	    }
	}

}

void
load_ID_table (uint8_t *buf, uint32_t rlen, uint32_t mach_type, uint32_t *IDtabsiz)
{
	dprintf (DEBUG_TABLES, "Copying IDtable of %d bytes to position %d\n", rlen, *IDtabsiz);
	memcpy (IDtable + *IDtabsiz, buf, rlen);
	*IDtabsiz += rlen;
	dprintf (DEBUG_TABLES, "load_ID_Table: NEW ID table size = %d\n", IDtablesize);
}

/*
 * ID FORMAT
 *
 * BIT  15   14            10 9                             0
 *      -----------------------------------------------------
 *      :NOT                                                :   
 *      :     IDENT CHARACTER        IDENTIFICATION NUMERAL :
 *      :USED                                               :
 *      -----------------------------------------------------
 *
 *        The IDENT character is an ASCII equivalent of 'A' thru 'Z' minus 100(8).
 *        The IDENTIFICATION NUMERAL is a range from 0 to 999(10).
 *        I.e. A000 through Z999.
 */

void
dump_ID_table (uint8_t * buf, uint32_t length, uint32_t mach_type)
{
	uint32_t	uid;
	int		i, j;
	int		first;

	char	*ID_caps[] = { "?15", "RP", "JT", "JL", "JP", "JI",
				"JM", "MT", "PP", "PR", "LP", "CR",
				"XX", "PFA", "FCP", "MWA" };
#ifdef NOTYET
	char	*ID_caps_desc[] = {
			"Undefined bit 15",
			"Reader Punch",
			"Job Transmitter",
			"Job Line Printer",
			"Job Punch",
			"Job Inquiry",
			"Job Message",
			"Magnetic Tape Unit",
			"Paper Tape Punch",
			"Paper Tape Reader",
			"Line Printer",
			"Card Reader",
			"Undefined2",
			"Program/File Access",
			"File Create/Purge",
			"Multi Write Access"
	};
#endif

	eprintf ("ID Table:\n");

	eprintf ("                      Time         Disc\n");
	eprintf ("    ID    Passwd  Allow/Used   Allow/Used");
	if (MACH_TYPE_ACCESS(mach_type)) {
	eprintf ("   Capabilities   Device Designators\n");
		for (i = 0; i < length / 24; i++, buf += 24) {
			uid = WTOI (buf);
			eprintf ("    %c%03d  %-6.6s  %.5d/%.5d  %.5d/%.5d",
					(uid / 1024) + '@', uid & 0x3ff, 
					printarrow(line, buf + 2, 6, 0),
					WTOI (buf + 8),
					WTOI (buf + 10),
					WTOI (buf + 12),
					WTOI (buf + 14));
			first = 0;
			for (j = 1; j < 4; j++) {
				if (!first++)
					eprintf (" ");
				if (WTOI (buf + 16) & (1 << (j - 1)))
					eprintf (" %3.3s", ID_caps[j + 12]);
				else
					eprintf ("    ");
			}
			first = 0;
			for (j = 15; j > 3 ; j--) {
				if (WTOI (buf + 16) & (1 << (j - 1))) {
					eprintf ("%s%2.2s", !first ? "    " : ",", ID_caps[16 - j]);
					first ++;
				}
			}
			eprintf ("\n");
		}
	} else if (MACH_TYPE_2000F(mach_type)) {
		eprintf ("\n");
		buf += 2;
		for (i = 0; i < length / 16; i++, buf += 16) {
			uid = WTOI (buf);
			eprintf ("    %c%03d  %-6.6s  %05d/%05d  %05d/%05d\n",
					(uid / 1024) + '@', uid & 0x3ff, 
					printarrow(line, buf + 2, 6, 0),
					WTOI (buf + 8),
					WTOI (buf + 10),
					WTOI (buf + 12),
					WTOI (buf + 14));
		}
	}
}

uint8_t		* ptr, * ptrend;

uint8_t	*
list_opair (uint8_t *line, uint32_t firstop, uint32_t opair, uint8_t * ptr)
{
	uint32_t	opr, opd;
	uint32_t	op15, opn;
	uint8_t	*	ptr_start = ptr;
	uint32_t	i;

	dprintf (DEBUG_DETOKEN2, ".");
	opair = WTOI (ptr);
	opr = (opair >> 9) & 077;
	opd = opair & 0377;
	op15 = opair >> 15;

	if (firstop && MACH_TYPE_ACCESS(tape_mach_type)) {
		if (opr != 073 && opr != 001) {
			dprintf (DEBUG_INDCONV, "%s ", Bst_ACCESS[opr].ascii);
			lprintf (line, "%s ", Bst_ACCESS[opr].ascii);
		}
	} else {
		dprintf (DEBUG_DETOKEN2, "~%d~", opr);
		if (opr != 001) {
			switch (tape_mach_type) {
			case MACH_TYPE_SET_ACCESSA:
			case MACH_TYPE_SET_ACCESSB:
				dprintf (DEBUG_INDCONV, "%s", Bop_ACCESS[opr].ascii);
				lprintf (line, "%s", Bop_ACCESS[opr].ascii);
				break;
			case MACH_TYPE_SET_2000F:
			case MACH_TYPE_SET_2000FP:
				dprintf (DEBUG_INDCONV, "%s", Bop_2000F[opr].ascii);
				lprintf (line, "%s", Bop_2000F[opr].ascii);
				break;
			case MACH_TYPE_SET_2000E:
				dprintf (DEBUG_INDCONV, "%s", Bop_2000E[opr].ascii);
				lprintf (line, "%s", Bop_2000E[opr].ascii);
				break;
			case MACH_TYPE_SET_2000C:
			case MACH_TYPE_SET_2000CP:
				dprintf (DEBUG_INDCONV, "%s", Bop_2000C[opr].ascii);
				lprintf (line, "%s", Bop_2000C[opr].ascii);
				break;
			case MACH_TYPE_SET_2000B:
				dprintf (DEBUG_INDCONV, "%s", Bop_2000B[opr].ascii);
				lprintf (line, "%s", Bop_2000B[opr].ascii);
				break;
			case MACH_TYPE_SET_2000A:
				dprintf (DEBUG_INDCONV, "%s", Bop_2000A[opr].ascii);
				lprintf (line, "%s", Bop_2000A[opr].ascii);
				break;
			default:	/* Wing it and assume an HP2000F syntax */
				dprintf (DEBUG_INDCONV, "%s", Bop_2000F[opr].ascii);
				lprintf (line, "%s", Bop_2000F[opr].ascii);
				break;
			}
		}
	}

	if (opr == 1) {			/* String constant */
		int i;

		ptr++;
		quoteflag = 0;
		if (opd == 0) {
			dprintf (DEBUG_INDCONV, "\"\"");
			lprintf (line, "\"\"");
		} else {
			for (i = 0; i < opd; i++) {
				HPputchar (line + strlen(line), ++ptr, 0, &quoteflag);
			}
		}
		if (quoteflag) {
			dprintf (DEBUG_INDCONV, "\"");
			lprintf (line, "\"");
		}
		ptr += 2;
		if (!(opd % 2))
			ptr--;
		dprintf (DEBUG_DETOKEN2, "{");
		return (ptr);
	}
	dprintf (DEBUG_DETOKEN, " >>> opair=%d op15=%d opd=%d\n", opair, op15, opd);

	if (opd == 132 && MACH_TYPE_2000E(tape_mach_type)) {	/* Hack for 2000E FNA(X) to force X output */
		dprintf (DEBUG_DETOKEN, " NEWopd >>> opair=%d op15=%d opd=%d\n", opair, op15, opd);
		op15 = 0;
	}
	if (!op15 && !opd && !opr) {
		ptr += 2;
		dprintf (DEBUG_DETOKEN2, "}");
		return (ptr);
	}
	opd = (opair & 017);
	dprintf (DEBUG_DETOKEN, " NEWopd >>> opair=%d op15=%d opd=%d\n", opair, op15, opd);
	if (op15) {
		opn = (opair >> 4) & 037;
		if (!opd) {
			char  valconv[25];
			ptr += 2;
			dprintf (DEBUG_INDCONV, "%s", ftoHP(HPftoIEEE( (WTOI(ptr) << 16) + WTOI(ptr + 2)), valconv));
			lprintf (line, "%s", ftoHP(HPftoIEEE( (WTOI(ptr) << 16) + WTOI(ptr + 2)), valconv));
			ptr += 4;
		} else if (opd == 3) {
			uint32_t	newline;
			ptr += 2;
			if (opr == 074) {			/* GOTO/GOSUB OF */
				i = 0;
				while (ptr < ptrend) {
					newline = WTOI (ptr);
					if (ENTRY_TYPE_CSAVE(Pentry.type)) {
						if (newline > Pentry.startloc
								&& newline < Pentry.startloc + Pentry.symtabstart)
							newline = WTOI (Pentry.block + ((newline - Pentry.startloc) * 2));
						else
							newline = 0;
					}
					dprintf (DEBUG_INDCONV, "%s%d", i ? "," : "", newline);
					lprintf (line, "%s%d", i ? "," : "", newline);
					ptr += 2;
					i++;
				}
			} else {				/* Simple parameter */
				newline = WTOI (ptr);
				if (ENTRY_TYPE_CSAVE(Pentry.type) &&
						(opr == 043 ||	/* USING */
						 opr == 052 ||	/* GOTO */
						 opr == 056 ||	/* GOSUB */
						 opr == 066 ||	/* RESTORE */
						 opr == 075))	/* THEN */
					newline = WTOI (Pentry.block + ((newline - Pentry.startloc) * 2));
				dprintf (DEBUG_INDCONV, "%d", newline);
				lprintf (line, "%d", newline);
				ptr += 2;
			}
		} else if (opd == 017) {			/* Built-in function */
			dprintf (DEBUG_INDCONV, "%s", Bfunc[opn].ascii);
			lprintf (line, "%s", Bfunc[opn].ascii);
			ptr += 2;
		} else {
			dprintf (DEBUG_INDCONV, "%c", 'A' + opn - 1);		/* Simple variable in DEF FN */
			lprintf (line, "%c", 'A' + opn - 1);
			ptr += 2;
		}
	} else {
		opn = (opair >> 4) & 037;
		if (ENTRY_TYPE_CSAVE(Pentry.type)) {		/* Re-lookup variable in symbol table */
			if (Pentry.symtabstart + Pentry.block < Pentry.block + Pentry.filesize) {
				opair = WTOI (Pentry.block + Pentry.symtabstart + ((opd - 1 + (opn << 4)) * 4));
				opn = (opair >> 4) & 037;
				opd = opair & 017;
			}
		}
		if (opn > 032 && opn < 037) {			/* A1$ variable */
			dprintf (DEBUG_INDCONV, "%c%d$", 'A' + 16 * ((opn - 1) % 2) + opd, (opn - 033) / 2);
			lprintf (line, "%c%d$", 'A' + 16 * ((opn - 1) % 2) + opd, (opn - 033) / 2);
		} else if (opd == 0) {				/* A$ variable */
			if (opn != 0) {
				dprintf (DEBUG_INDCONV, "%c$", 'A' + opn - 1);
				lprintf (line, "%c$", 'A' + opn - 1);
			}
		} else if (opd >= 1 && opd <= 4) {		/* Dimensioned variable */
			if (opn != 0) {
				dprintf (DEBUG_INDCONV, "%c", 'A' + opn - 1);
				lprintf (line, "%c", 'A' + opn - 1);
			}
		} else if (opd == 017) {			/* User-defined function */
			dprintf (DEBUG_INDCONV, "FN%c", 'A' + opn - 1);
			lprintf (line, "FN%c", 'A' + opn - 1);
		} else if (opd >= 5 && opd <= 16) {		/* A1 variable */
			dprintf (DEBUG_INDCONV, "%c%d", 'A' + opn - 1, opd - 5);
			lprintf (line, "%c%d", 'A' + opn - 1, opd - 5);
		} else {
			dprintf (DEBUG_INDCONV, "%c", 'A' + opn - 1);		/* Simple variable */
			lprintf (line, "%c", 'A' + opn - 1);
		}
		ptr += 2;
	}
	dprintf (DEBUG_DETOKEN, "LEAVING3 list_opair: ptr=%d ptrend=%d\n", ptr, ptrend);
	if (ptr_start == ptr) {
		dprintf (DEBUG_INDCONV, "***** ptr did not increment -- forcing\n");
		ptr += 2;
	}
	dprintf (DEBUG_DETOKEN2, "\\");
	return (ptr);
}

void
convert_asciifile (struct Program_Entry *Pentry)
{
	uint32_t	recno, stlen, astlen;
	uint8_t		line[520];
	uint32_t	len, nlen;
	struct line_table *ltab;

	recno = 0;
	ptr = Pentry->block;
	ptrend = ptr;
	len = 0;
	Pentry->cline.len = 0;
	Pentry->cline.data = NULL;
	Pentry->cline.next = NULL;
	ltab = &Pentry->cline;

	if (Pentry->block_size == 0)
		return;
	dprintf (DEBUG_FILES, "convert_ascii: Pentry->startloc=%d  Pentry->len=%d\n",
			Pentry->startloc, Pentry->len);
	dprintf (DEBUG_FILES, "ASCII FILE... ptr=%d Pentry->filesize=%d\n", Pentry->block, Pentry->block_size);
	while (ptr < Pentry->block + Pentry->block_size && WTOI(ptr) != 0xffff) {
		recno ++;
		stlen = WTOI(ptr);
		astlen = stlen + ((stlen % 2) == 1);
		ptrend = ptr + 4 + astlen;
		if (verbose > 5) {
			eprintf("ASCII RECORD IS:\n");
			hexdump(ptr, 10);
			eprintf("---> ptr=%d  stlen=%d  astlen=%d  ptrend=%d\n",
					ptr, stlen, astlen, ptrend);
		}
		ptr += 2;
		if (stlen == 65535)
			return;
		if (stlen == 65534) {
			dprintf (DEBUG_FILES, "ASCII EOR: SKIPPING %d BYTES\n",
				512 - (ptr - Pentry->block) % 512);
			ptr = ptr + (512 - (ptr - Pentry->block) % 512);
			continue;
		}
		if (astlen > 1000)
			errorexit (-99, "ASCII file corrupt--TOO MUCH DATA. astlen=%d\n",
					astlen);
			
		dprintf (DEBUG_FILES, "ASCII---> recno=%d  ptr=%d  stlen=%d  astlen=%d  ptrend=%d\n",
				recno, ptr, stlen, astlen, ptrend);
		
		nlen = (uint32_t) (printarrow (line, ptr, stlen, 1));
		ptr += astlen;
		if (convert || display)
			add_line (&ltab, nlen, line);
	}
}

void
convert_file (struct Program_Entry *Pentry)
{
	uint32_t	recno, stlen, astlen;
	uint32_t	first = 0;
	uint8_t		*iptr;
	uint8_t		line[512 * 1024];
	uint32_t	quoteflag;
	uint8_t		valconv[25];
	uint32_t	skip;
	uint32_t	linelen = 0;
	struct line_table *ltab;

	Pentry->cline.len = 0;
	Pentry->cline.data = NULL;
	Pentry->cline.next = NULL;
	ltab = &Pentry->cline;

	recno = 0;
	ptr = Pentry->block;
	ptrend = ptr;
	line[0] = 0;

	if (Pentry->block_size == 0)
		return;
	if (Pentry->startloc == 0) {
		eprintf ("File entry has a startloc of zero -- Can't process.\n");
		return;
	}
	dprintf (DEBUG_FILES, "STANDARD FILE: filesize=%d  block_size=%d  startloc=%d  len=%d\n",
			Pentry->filesize, Pentry->block_size, Pentry->startloc, Pentry->len);
	recno++;
	while (ptr < Pentry->block + Pentry->block_size) {
		skip = Pentry->startloc * 2 - ((ptr + 2 - Pentry->block) % (Pentry->startloc * 2));
		if (skip == (Pentry->startloc * 2) - 2) {
			first = 0;
			dprintf (DEBUG_FILES, "\nRecord #%d:\n", recno);
			if (linelen && (convert || display)) {
				dprintf (DEBUG_FILES, "Adding line: %.*s\n",
						linelen, line);
				add_line (&ltab, linelen, line);
			}
			linelen = 0;
		}
		stlen = WTOI(ptr);
		if (verbose > 5) {
			eprintf("FILE RECORD IS (first 10 bytes):\n");
			hexdump(ptr, 10);
			eprintf("---> ptr=%d  stlen=%d\n", ptr, stlen);
		}
		if (stlen == 65535) {
			if (first) {
				if (convert || display) {
					dprintf (DEBUG_FILES, "Adding line: %.*s\n",
							linelen, line);
					add_line (&ltab, linelen, line);
				}
				linelen = 0;
			}
			recno++;
			first = 0;
			ptr += 2;
			dprintf (DEBUG_FILES, "\nFILE EOR: SKIPPING %d BYTES\n", skip);
			if (verbose > 5)
				hexdump (ptr, skip);
			ptr = ptr + skip;
			continue;
		}
		if (stlen == 65534) {
			if (first) {
				if (convert || display) {
					dprintf (DEBUG_FILES, "Adding line: %.*s\n",
							linelen, line);
					add_line (&ltab, linelen, line);
				}
				linelen = 0;
			}
			recno++;
			first = 0;
			ptr += 2;
			dprintf (DEBUG_FILES, "\nFILE EOR: SKIPPING %d BYTES\n", skip);
			if (verbose > 5)
				hexdump (ptr, skip);
			ptr = ptr + skip;
			continue;
		}
		if (first)
			line[linelen++] = ',';
		else
			first++;

		if ((stlen & 0xff00) != 0x0200) {			/* Floating-point number */
			sprintf (line + linelen, "%s",
				ftoHP(HPftoIEEE( (WTOI(ptr) << 16) + WTOI(ptr + 2)), valconv));
			linelen += strlen(valconv);
			ptr += 4;
			continue;
		}
		stlen = stlen & 0xff;
		astlen = stlen + ((stlen % 2) == 1);
		dprintf (DEBUG_FILES, "ASCII STRING LENGTH: %d (%x) of %d\n", stlen, stlen, astlen);
		if (astlen > 1000)
			errorexit (-99, "Regular file corrupt--TOO MUCH DATA! astlen=%d\n",
					astlen);
		ptr += 2;
		ptrend = ptr + 2 + astlen;
			
		dprintf (DEBUG_FILES, "ASCII---> recno=%d  ptr=%d  stlen=%d  astlen=%d  ptrend=%d\n",
				recno, ptr, stlen, astlen, ptrend);
		line [linelen++] = '"';
		for (iptr = ptr; iptr < ptr + stlen; iptr++)
			linelen += HPputchar (line + linelen , iptr, 1, &quoteflag);
		line [linelen++] = '"';
		ptr += astlen;
	}
}

void
convert_program (struct Program_Entry  * Pentry)
{
	uint8_t		line[32768];
	uint32_t	lineno;
	uint32_t	stlen;
	uint32_t	opair, opr, opd;
	uint32_t	op15;
	struct line_table *ltab;

	Pentry->cline.len = 0;
	Pentry->cline.data = NULL;
	Pentry->cline.next = NULL;
	ltab = &Pentry->cline;

	ptr = Pentry->block;
	ptrend = ptr;

	if (ENTRY_TYPE_CSAVE(Pentry->type))
		dprintf (DEBUG_DETOKEN2, "------> SEMI-COMPILED PROGRAM\n");
	dprintf (DEBUG_DETOKEN2, "------> Startloc: %d  Relativestartloc: %d\n",
			Pentry->startloc, Pentry->startloc - 02000);
	dprintf (DEBUG_DETOKEN2, "------> filesize: %d  symtabstart: %d\n",
			Pentry->filesize, Pentry->symtabstart);

	while (ptr < Pentry->block + (ENTRY_TYPE_CSAVE(Pentry->type)
			? Pentry->symtabstart : Pentry->filesize)) {

		if (display && line[0])
			dprintf (DEBUG_INDCONV, "C> (len=%d) %s\n", strlen(line), line);
		if ((display || convert) && line[0])
			add_line (&ltab, strlen(line), line);

		dprintf (DEBUG_CONVERT, "\n");
		dprintf (DEBUG_CONVERT, ".... NEXT STATEMENT START -----> ptr=%d -- MOD=%d\n",
				ptr, ptr - Pentry->block);
		lineno = WTOI (ptr);
		line[0] = 0;
		ptr += 2;
		dprintf (DEBUG_CONVERT, "LINE NO: %d", lineno);

		stlen = WTOI (ptr);
		ptrend = ptr + (stlen * 2) - 2;
		ptr += 2;

		dprintf (DEBUG_CONVERT, "           STATEMENT LEN: %d\n", stlen);

		if (ptr + (stlen - 2) * 2 > ptrend) {
			dprintf (DEBUG_INDCONV, "File %.6s corrupt -- Statement length "
					"exceeds data available by %d bytes.\n",
					Pentry->filnam, ptrend - (ptr + (stlen - 1) * 2));
			dprintf (DEBUG_INDCONV, "      stlen=%d  ptr=%d  ptrend=%d\n",
					stlen, ptr, ptrend);
			return;
		}

		opair = WTOI (ptr);
	
		opr = (opair >> 9) & 077;
		opd = opair & 0377;
		op15 = opair >> 15;

		if (opr == 051) {		/* REM */
			ptr += 1;
			dprintf (DEBUG_INDCONV, "%d  %s%.*s\n",
					lineno, Bst_ACCESS[opr].ascii, ptrend - ptr, ptr);
			lprintf (line, "%d  %s%.*s",
					lineno, Bst_ACCESS[opr].ascii, ptrend - ptr, ptr);
			ptr = ptrend;
			continue;
		} else if (opr == 070) {	/* FILES */
			ptr += 1;
			dprintf (DEBUG_INDCONV, "%d  %s %.*s\n",
					lineno, Bst_ACCESS[opr].ascii, ptrend - ptr, ptr);
			lprintf (line, "%d  %s %.*s",
					lineno, Bst_ACCESS[opr].ascii, ptrend - ptr, ptr);
			ptr = ptrend;
			continue;
		} else if (opr == 044) {	/* IMAGE */
			ptr += 2;
			quoteflag = 0;
			dprintf (DEBUG_INDCONV, "%d  %s%.*s\n",
					lineno, Bst_ACCESS[opr].ascii, opair & 0xff, ptr);
			lprintf (line, "%d  %s%.*s",
					lineno, Bst_ACCESS[opr].ascii, opair & 0xff, ptr);
			ptr = ptrend;
			continue;
		} else {
			dprintf (DEBUG_DETOKEN, "	calling list_opair: ptr=%d  ptrend=%d\n",
					ptr, ptrend);
			if (MACH_TYPE_ACCESS(tape_mach_type)) {
				dprintf (DEBUG_INDCONV, "%d  ", lineno);
				lprintf (line, "%d  ", lineno);
			} else {
				dprintf (DEBUG_INDCONV, "%d ", lineno);
				lprintf (line, "%d ", lineno);
			}
			ptr = list_opair (line, 1, opair, ptr);
			while (ptr < ptrend) {
				dprintf (DEBUG_DETOKEN, "	calling list_opair: ptr=%d  ptrend=%d\n",
						ptr, ptrend);
				ptr = list_opair (line, 0, opair, ptr);
			}
		}
		dprintf (DEBUG_INDCONV, "\n");
		dprintf (DEBUG_INDCONV, "||| %s\n", line);
	}
	if (display && line[0])
		dprintf (DEBUG_INDCONV, "C> %s\n", line);
	if ((display || convert) && line[0])
		add_line (&ltab, strlen(line), line);
	line[0] = 0;
}

/* TEMP */
uint32_t	lrlen;
	
void
load_program_entry (uint32_t program_entry_state, uint8_t *buf, uint32_t rlen, uint32_t mach_type)
{
	uint8_t		entry[12];
	int		i;

	switch (program_entry_state) {
	case 0:	
		memset (&Pentry, 0, sizeof (Pentry));

		lrlen = (rlen - 24);

		if (MACH_TYPE_ACCESS(mach_type)) {

			memcpy (Pentry.header, buf, 24);

			Pentry.id = WTOI (buf);
			memcpy (Pentry.HPfilnam, buf + 2, 6);
			Pentry.startloc = WTOI (buf + 8);
			Pentry.refdate = WTOI (buf + 10);
			Pentry.moddate = WTOI (buf + 12);
			Pentry.status = WTOI (buf + 14);
			Pentry.daddr = WTOI (buf + 16);
			Pentry.ddev = WTOI (buf + 18);
			Pentry.linfo = WTOI (buf + 20);
			Pentry.len = WTOI (buf + 22);

			if (Pentry.HPfilnam[4] & 0x80) {
				Pentry.type = ENTRY_TYPE_SET_CSAVE;
				Pentry.HPfilnam[4] &= 0x7f;
				dprintf (DEBUG_FILEREAD, "1-type CSAVE\n");
			}
			if (Pentry.HPfilnam[2] & 0x80) {
				Pentry.type = ENTRY_TYPE_SET_FILE;
				Pentry.HPfilnam[2] &= 0x7f;
				dprintf (DEBUG_FILEREAD, "1-type FILE\n");
			}
			if (Pentry.HPfilnam[0] & 0x80) {
				Pentry.type = ENTRY_TYPE_SET_ASCII;
				Pentry.HPfilnam[0] &= 0x7f;
				dprintf (DEBUG_FILEREAD, "1-type ASCII\n");
			}
			if (Pentry.type == 0)
				dprintf (DEBUG_FILEREAD, "1-type PROGRAM\n");

			if (ENTRY_TYPE_PROGRAM(Pentry.type)) {
				Pentry.filesize = (65536 - Pentry.len) * 2;
				Pentry.block_count = (Pentry.filesize + 2047) / 2048;
			} else if (ENTRY_TYPE_FILE(Pentry.type)) {
				Pentry.filesize = (Pentry.len * 256) * 2;
				Pentry.block_count = (Pentry.filesize + (Pentry.startloc - 1)) / 2048;
			} else if (ENTRY_TYPE_ASCII(Pentry.type)) {
				Pentry.filesize = (Pentry.len * 256) * 2;
				Pentry.block_count = Pentry.filesize / 2048;
			}

			dprintf (DEBUG_FILEREAD, "1-Reading file %6.6s: startloc=%d refdate=%d moddate=%d "
					"status=%d daddr=%d ddev=%d linfo=%d len=%d "
					"block_count=%d block_size=%d type=%d filesize=%d\n",
					Pentry.HPfilnam, Pentry.startloc, Pentry.refdate, 
					Pentry.moddate, Pentry.status, Pentry.daddr, Pentry.ddev, 
					Pentry.linfo, Pentry.len, Pentry.block_count,
					Pentry.block_size, Pentry.type, Pentry.filesize);

			dprintf (DEBUG_FILEREAD, "1-Allocating %d bytes for Pentry.block\n", Pentry.filesize);
			Pentry.block = calloc(Pentry.filesize * 2, sizeof (uint32_t));
			if (Pentry.block == NULL)
				errorexit (-9, "Can't allocate %d bytes for block_count: %s\n",
						Pentry.block_size, strerror (errno));

			/*
			 * Accomodate minimum 24-byte tape-record length.  If the entry is less than
			 * 24 bytes, only copy the actual bytes needed (the filesize count itself).
			 */
			lrlen = rlen - 24;
			if (!ENTRY_TYPE_ASCII(Pentry.type)
					&& !ENTRY_TYPE_FILE(Pentry.type)
					&& Pentry.filesize < 24) {
				lrlen = Pentry.filesize;
				dprintf (DEBUG_FILEREAD, "Adjusting record size down from %d to %d"
						" for short-length record minimum.\n",
					rlen - 24, lrlen);
			}

			dprintf (DEBUG_FILEREAD, "1-Copying %d bytes from buf to Pentry.block\n",
					lrlen);
			memcpy (Pentry.block + Pentry.block_size, buf + 24, lrlen);
			Pentry.block_size += lrlen;

			dprintf (DEBUG_FILEREAD, "2-Reading file %6.6s: startloc=%d refdate=%d "
					"moddate=%d status=%d daddr=%d ddev=%d "
					"linfo=%d len=%d block_count=%d "
					"block_size=%d type=%d filesize=%d\n",
					Pentry.HPfilnam, Pentry.startloc,
					Pentry.refdate, Pentry.moddate,
					Pentry.status, Pentry.daddr, 
					Pentry.ddev, Pentry.linfo,
					Pentry.len, Pentry.block_count,
					Pentry.block_size, Pentry.type,
					Pentry.filesize);

		} else if (MACH_TYPE_2000F(mach_type) || MACH_TYPE_2000C(mach_type)) {

			memcpy (Pentry.header, buf + 2, 24);

			Pentry.id = WTOI (buf + 2);
			memcpy (Pentry.HPfilnam, buf + 4, 6);
			Pentry.startloc = WTOI (buf + 10);
			Pentry.refdate = WTOI (buf + 12);
			Pentry.moddate = WTOI (buf + 14);
			Pentry.drumaddr = WTOI (buf + 16);
			Pentry.daddr = WTOI (buf + 18);
			Pentry.ddev = WTOI (buf + 20);
			Pentry.linfo = WTOI (buf + 22);
			Pentry.len = WTOI (buf + 24);

			if (Pentry.HPfilnam[0] & 0x80) {
				Pentry.type |= ENTRY_TYPE_SET_PROT;
				Pentry.HPfilnam[0] &= 0x7f;
				dprintf (DEBUG_INDCONV, "1-TYPE PROT\n");
			}
			if (Pentry.HPfilnam[4] & 0x80) {
				Pentry.type |= ENTRY_TYPE_SET_CSAVE;
				Pentry.HPfilnam[4] &= 0x7f;
				dprintf (DEBUG_INDCONV, "1-type CSAVE\n");
			}
			if (Pentry.HPfilnam[2] & 0x80) {
				Pentry.type = (Pentry.type | ENTRY_TYPE_SET_FILE)
						& (~ENTRY_TYPE_SET_CSAVE);
				Pentry.HPfilnam[2] &= 0x7f;
				dprintf (DEBUG_INDCONV, "1-type FILE\n");
			}
			if (Pentry.type == 0)
				dprintf (DEBUG_INDCONV, "1-type PROGRAM\n");

			if (ENTRY_TYPE_PROGRAM(Pentry.type)) {
				Pentry.filesize = (65536 - Pentry.len) * 2;
				Pentry.block_count = (Pentry.filesize + 2047) / 2048;
			} else if (ENTRY_TYPE_FILE(Pentry.type)) {
				Pentry.filesize = (Pentry.len * 256) * 2;
				Pentry.block_count = (Pentry.filesize + 
						(Pentry.startloc - 1)) / 2048;
			}

			dprintf (DEBUG_FILEREAD, "1-Reading file %6.6s: startloc=%d refdate=%d moddate=%d "
					"status=%d daddr=%d ddev=%d linfo=%d len=%d "
					"block_count=%d block_size=%d type=%d filesize=%d\n",
					Pentry.HPfilnam, Pentry.startloc, Pentry.refdate, 
					Pentry.moddate, Pentry.status, Pentry.daddr, Pentry.ddev, 
					Pentry.linfo, Pentry.len, Pentry.block_count,
					Pentry.block_size, Pentry.type, Pentry.filesize);

			dprintf (DEBUG_FILEREAD, "1-Allocating %d bytes for Pentry.block\n", Pentry.filesize);
			Pentry.block = calloc(Pentry.filesize * 2, sizeof (uint8_t));
			if (Pentry.block == NULL)
				errorexit (-9, "Can't allocate %d bytes for block_count: %s\n",
						Pentry.block_size, strerror (errno));

			/*
			 * Accomodate for minimum 24-byte tape-record length.  If the entry is less than 24 bytes,
			 * only copy the actual bytes needed (the filesize count itself).
			 */
			lrlen = rlen  - 26;
			if (Pentry.filesize < 24) {
				lrlen = Pentry.filesize;
				dprintf (DEBUG_FILEREAD, "Adjusting record size down from %d to %d for short-length record minimum.\n",
					rlen - 26, lrlen);
			}

			dprintf (DEBUG_FILEREAD, "1-Copying %d bytes from buf to Pentry.block\n", lrlen);
			memcpy (Pentry.block + Pentry.block_size, buf + 26, lrlen);
			Pentry.block_size += lrlen;

			dprintf (DEBUG_FILEREAD, "2-Reading file %6.6s: startloc=%d refdate=%d moddate=%d "
					"status=%d daddr=%d ddev=%d linfo=%d len=%d block_count=%d "
					"block_size=%d type=%d filesize=%d\n",
					Pentry.HPfilnam, Pentry.startloc, Pentry.refdate, 
					Pentry.moddate, Pentry.status, Pentry.daddr, 
					Pentry.ddev, Pentry.linfo, Pentry.len, Pentry.block_count,
					Pentry.block_size, Pentry.type, Pentry.filesize);

		} else if (MACH_TYPE_2000A(mach_type) || MACH_TYPE_2000B(mach_type) ||
				MACH_TYPE_2000E(mach_type)) {
			Pentry.id = WTOI (buf);
			memcpy (Pentry.HPfilnam, buf+2, 6);
			Pentry.refdate = WTOI (buf + 10);
			Pentry.daddr = WTOI (buf + 12);
			Pentry.len = WTOI (buf + 14);
		} else {
			dprintf (DEBUG_HEADERS, "DIRENT: Can't accomodate Directory Entries for unknown "
					"machine types.\n");
			return;
		}
		memcpy (Pentry.filnam, Pentry.HPfilnam, 6);
		Pentry.filnam[6] = 0;

		/* Look for the first non-black character from the end of the filename - It's the end */
		for (i = 6; i > 0; i--)
			if (Pentry.filnam[i - 1] != ' ')
				break;
		Pentry.filnam[i] = 0;
		/* There may be embedded spaces in the filename.  Change 'em to '/' */
		for (i = 0; i < strlen(Pentry.filnam); i++)
			if (Pentry.filnam[i] == ' ')
				Pentry.filnam[i] = '/';

#ifdef OLD
		if (strchr(Pentry.filnam, ' ') != NULL)
			*(strchr(Pentry.filnam, ' ')) = 0;
#endif
		
		(void) sprintf (entry, "%s.%c%03d", Pentry.filnam,
					Pentry.id / 1024 + '@',
					Pentry.id % 1024);
		if (check_entrylist(entry, rentry[0], 12, nentries)) {
			if (verbose > 1) {
				if (Pentry.id != last_progID)
					eprintf ("%c%03d     %.6s     %03d/%02d     ",
							Pentry.id / 1024 + '@',
							Pentry.id % 1024,
							Pentry.HPfilnam,
							Pentry.refdate & 0x1ff,
							Pentry.refdate >> 9);
				else
					eprintf ("         %.6s     %03d/%02d     ",
							Pentry.HPfilnam,
							Pentry.refdate & 0x1ff,
							Pentry.refdate >> 9);
				last_progID = Pentry.id;
				if (MACH_TYPE_ACCESS(mach_type)) {
					switch (Pentry.type) {
					case ENTRY_TYPE_SET_ASCII:
						eprintf ("A");
						break;
					case ENTRY_TYPE_SET_FILE:
						if (ENTRY_TYPE_FILE(Pentry.type) && Pentry.len == 0)
							eprintf ("A");
						else
							eprintf ("F");
						break;
					case ENTRY_TYPE_SET_CSAVE:
						eprintf ("C");
						break;
					default:
						eprintf (" ");
					}
					if (Pentry.status & 0x01)
						eprintf ("U");
					else if (Pentry.status & 0x02)
						eprintf ("P");
					else if (Pentry.status & 0x04)
						eprintf ("L");
					else if (Pentry.status & 0x08)
						eprintf (" ");
					else
						eprintf ("X");
				} else if (MACH_TYPE_2000F(mach_type)) {
					if (ENTRY_TYPE_PROT(Pentry.type))
						eprintf ("P");
					else if (ENTRY_TYPE_FILE(Pentry.type))
						eprintf ("F");
					else if (ENTRY_TYPE_CSAVE(Pentry.type))
						eprintf ("C");
					else
						eprintf (" ");
				}
				eprintf ("     ");
				if (ENTRY_TYPE_FILE(Pentry.type) && Pentry.len == 0)
					eprintf ("          %c%c%d   ",
							(Pentry.ddev >> 10) + 0101,
							((Pentry.ddev>> 5) & 0x1f) + 0101,
							(Pentry.ddev & 0x1f) - 1);
				else
					eprintf ("%05d     %06d",
							(ENTRY_TYPE_FILE(Pentry.type)) ? Pentry.len
									: (65536 - Pentry.len + 255) / 256,
							Pentry.ddev);
				if (MACH_TYPE_ACCESS(mach_type) && ENTRY_TYPE_FILE(Pentry.type) && Pentry.startloc != 256) {
					eprintf ("     ");
					if (Pentry.len != 0)
						eprintf ("%04d", Pentry.startloc);
				}
				if ((MACH_TYPE_2000F(mach_type) || MACH_TYPE_2000C(mach_type))
								&& Pentry.drumaddr != 0)
					eprintf ("     %d\n", Pentry.drumaddr >> 5,
						      	(Pentry.drumaddr >> 5) & 0x1f,
						      	Pentry.drumaddr & 0x1f);
	
				eprintf ("\n");
			
			} else
				if (verbose == 1)
					eprintf ("%s.%c%03d\n", Pentry.filnam,
							Pentry.id / 1024 + '@',
							Pentry.id % 1024);
		}
		break;
	case 1:

		/*
		 * Accomodate minimum 24-byte tape-record length.  If the
		 * remaining bytes needed is less than 24 bytes, only copy
		 * the actual bytes needed (the filesize less bytes already
		 * copied).  Only 2000C', 2000F and ACCESS have 24-byte
		 * minimum tape record lengths.
		 */
		lrlen = rlen;
		if (!MACH_TYPE_2000E(tape_mach_type)
				&& Pentry.filesize - Pentry.block_size < 24) {
			lrlen = Pentry.filesize - Pentry.block_size
					+ (MACH_TYPE_2000F(tape_mach_type) * 2);
			dprintf (DEBUG_FILEREAD, "Adjusting record size down from %d to %d "
					"for short-length record minimum.\n",
					rlen, lrlen);
		}

		if (MACH_TYPE_2000F(tape_mach_type)) {
			lrlen = lrlen - 2;
			buf = buf + 2;
		}
		dprintf (DEBUG_FILEREAD, "1-Copying %d bytes from buf to Pentry.block\n",
				lrlen);
		memcpy (Pentry.block + Pentry.block_size, buf, lrlen);
		Pentry.block_size += lrlen;
		break;
	case 2:
		if (Pentry.filesize != Pentry.block_size) {
			dprintf (DEBUG_TABLES, "Entry of %d bytes mismatch %d read bytes.  "
						"Diff=%2d bytes\n",
					Pentry.filesize, Pentry.block_size,
					Pentry.block_size - Pentry.filesize);
		}

		if (ENTRY_TYPE_CSAVE(Pentry.type))
			if (MACH_TYPE_ACCESS(tape_mach_type))
				Pentry.symtabstart =
					(WTOI (Pentry.block + Pentry.filesize - 1 - 11)
						- Pentry.startloc) * 2;
			else
				Pentry.symtabstart =
					(WTOI (Pentry.block + Pentry.filesize - 1 - 13)
						- Pentry.startloc) * 2;
		else
			Pentry.symtabstart = Pentry.filesize;

		sprintf (entry, "%s.%c%03d", Pentry.filnam,
				Pentry.id / 1024 + '@',
				Pentry.id % 1024);
		if (check_entrylist(entry, rentry[0], 12, nentries)) {
			if ((display || convert) && Pentry.id / 1024  > 0) {
				if (ENTRY_TYPE_CSAVE(Pentry.type)
						|| ENTRY_TYPE_PROGRAM(Pentry.type))
					convert_program (&Pentry);
				else if (ENTRY_TYPE_ASCII(Pentry.type))
					convert_asciifile (&Pentry);
				else if (ENTRY_TYPE_FILE(Pentry.type))
					convert_file (&Pentry);
				if (verbose > 3 || display) {
					struct line_table *lptr = &Pentry.cline;
					uint32_t	rno = 1;
					dprintf (DEBUG_FILES, "lptr = %d\n", *lptr);
					while (lptr->next != NULL) {
						dprintf (DEBUG_TABLES, "T> ");
						if (ENTRY_TYPE_ASCII(Pentry.type) ||
								ENTRY_TYPE_FILE(Pentry.type))
							dprintf (DEBUG_SYSTEM, "Record %d:\n", rno++);
						eprintf ("%.*s\n", lptr->len, lptr->data);
						lptr = lptr->next;
					}
				}
			}
			if (extract)
				extract_entry (&Pentry);
		}
		if (&Pentry.cline != NULL) {
			struct line_table *lptr = &Pentry.cline, *optr;
			while (lptr != NULL) {
				if (lptr->data != NULL)
					nfree(lptr->data);
				optr = lptr;
				lptr = lptr->next;
				if (optr != &Pentry.cline) {
					if (optr != NULL)
						nfree(optr);
				}
			}
		}
		nfree (Pentry.block);
		break;
	}
}
			
void
extract_entry (struct Program_Entry *Pentry)
{
	int	Ofd, nbytes;
	uint8_t	Ofilename[512];
	uint8_t	fixfilename[512];
	int	errx;

	strcpy (fixfilename, Pentry->filnam);
	dprintf (DEBUG_FILES, "fixfilename: %s\n", fixfilename);
	for (nbytes = 0; nbytes < strlen(fixfilename); nbytes++)
		if (fixfilename[nbytes] == '/' || fixfilename[nbytes] == '\\' ||
				fixfilename[nbytes] == ' ')
			fixfilename[nbytes] = '-';
			
	if (writedir) {
		sprintf (Ofilename, "%c%03d",
				Pentry->id / 1024 + '@',
				Pentry->id % 1024);
		errx = mkdir (Ofilename, 0644);
		if (errx != 0 && errno != EEXIST)
			errorexit (-19, "Can't create directory %s to write file %s: %s\n",
					Ofilename,
					fixfilename,
					strerror(errno));
		sprintf (Ofilename, "%c%03d/%s",
				Pentry->id / 1024 + '@',
				Pentry->id % 1024,
				fixfilename);
	} else
		sprintf (Ofilename, "%s.%c%03d",
				fixfilename,
				Pentry->id / 1024 + '@',
				Pentry->id % 1024);

	dprintf (DEBUG_FILES, "Writing file: %s\n", Ofilename);
	Ofd = open (Ofilename, O_CREAT | O_WRONLY | (overwrite ? O_TRUNC : O_EXCL) , 0666);
	if (Ofd < 0)
		eprintf ("MAIN: Can't open output file %s: %s\n",
				Ofilename, strerror (errno));
	else if (convert) {
		struct line_table *lptr = &Pentry->cline;
		uint32_t        rno = 1;

		while (lptr->next != NULL) {
			dprintf (DEBUG_FILES, "Writing segment %d of %d bytes:\n", rno++, lptr->len);
			nbytes = write (Ofd, lptr->data, lptr->len);
			if (nbytes != lptr->len)
				errorexit (-22, "Can't write %d data bytes to file %s: %s\n",
						lptr->len, Ofilename, strerror(errno));
			write (Ofd, "\n", 1);
			lptr = lptr->next;
		}
		lptr->next = NULL;
	} else {
		nbytes = write (Ofd, Pentry->header, 24);
		if (nbytes != 24)
			errorexit (-21, "Can't write %d header bytes to file %s: %s\n",
					Pentry->filesize, Ofilename, strerror(errno));
		nbytes = write (Ofd, Pentry->block, Pentry->filesize);
		if (nbytes != Pentry->filesize)
			errorexit (-22, "Can't write %d data bytes to file %s: %s\n",
					Pentry->filesize, Ofilename, strerror(errno));
	}
	close (Ofd);
}
